"use strict"

const NumToHex = (num, pad) => num.toString(16).padStart(pad || 2, '0').toUpperCase();
const ArrToHexStr = (inArr) => {
    let arr = Array.from(inArr);
    return arr.map(num => NumToHex(num)).join(",");
}

var localSIG = 256;
const GetSig = () => {
    localSIG++;
    if (localSIG > 255) localSIG = 0;
    return localSIG;
}

class PacketSpinel97 {
    
    PRE = 0x2A;       // PRE Prefix '*'
    FRM = 97;         // FRM Format 'a' 0x61
    NUM = 0x0000;
    ADR = 0x00;
    SIG = 0x00;
    INST = 0x00;
    DATA = [];
    SUM = 0x00;
    CR = 0x0D;

    constructor(arg) {

        if (typeof(arg) === 'boolean') {
            this.Init(arg);
        } else if (Array.isArray(arg) || Buffer.isBuffer(arg)) {
            this.CreateFromArray(...arg);
        } else if ((typeof arg === 'object') && (arg !== null)) {
            this.CreateFromObject(arg);
        } else {
            this.Init(true);
            this.RecountNUM();
            this.RecountSUM();
        }

    }

    Init(full_clear) {
        if (full_clear === true) {
            this.ADR = 0xFE;     // broadcast
            this.SIG = GetSig();
            this.INST = 0xF3;    // INFO 
            this.DATA = [];
        }
    }

    GetNUM() {
        let num = 5;
        if (Buffer.isBuffer(this.DATA)) num += this.DATA.length;
        return num;
    }

    RecountNUM() {
        this.NUM = this.GetNUM();
    }

    GetSUM() {
        let crc = this.PRE + this.FRM + (this.NUM & 0xFF00) + (this.NUM & 0x00FF) + this.ADR + this.SIG + this.INST;
        if (Buffer.isBuffer(this.DATA)) for (const one of this.DATA) crc += one;
        return 0xFF - (crc % 0x100);
    }

    RecountSUM() {
        this.SUM = this.GetSUM();
    }

    Check(IgnoreSum) {
        let error = [];
        if (this.PRE !== 0x2A) error.push("PRE");
        if (this.FRM !== 97) error.push("FRM");
        if (this.CR !== 0x0D) error.push("CR");
        if (this.NUM !== this.GetNUM()) error.push("NUM");
        if (IgnoreSum !== true) {
            let sum = this.GetSUM();
            if (this.SUM !== sum) {
                this.SUMok = sum;
                error.push("SUM");
            }
        }
        return (error.length > 0) ? error : null;
    }

    /**
     * Create Spinel packet from object
     * @param  {object} arg The object containing the individual parts of the packet.
     */
     CreateFromObject(arg) {
        this.ADR = Number.isInteger(arg.ADR) ? arg.ADR : 0xFE;
        this.INST = Number.isInteger(arg.INST) ? arg.INST : 0xF3;
        this.SIG = Number.isInteger(arg.SIG) ? arg.SIG : GetSig();
        this.DATA = (arg.DATA) ? Buffer.from(arg.DATA) : [];
        this.RecountNUM();
        this.RecountSUM();
    }

    /**
     * Create Spinel packet from variable array of arguments
     * @param  {...any} args Length: 0=Info request, 1=INST, 2=ADR+INST, 3=ADR+INST+SIG, Other=ADR+INST+SIG+[DATA]
     */
     CreateFromArray(...args) {
        this.Init(true);
        switch (args.length) {
            case 0:         // Info request
                this.INST = 0xF3;   
                break;
            case 1:         // INST
                this.INST = args[0];
                break;
            case 2:         // ADR, INST
                this.ADR = args[0];
                this.INST = args[1];
                break;
            case 3:         // ADR, INST, SIG
                this.ADR = args[0];
                this.INST = args[1];
                this.SIG = args[2];
                break;
            default:         // ADR, INST, SIG, [DATA]
                this.ADR = args[0];
                this.INST = args[1];
                this.SIG = args[2];
                this.DATA = Buffer.from(args[3]);
                break;
        }
        this.RecountNUM();
        this.RecountSUM();
    }

    GetBuffer(recount) {
        if (recount !== false) {            // false == data as filled in, without recalculations of NUM and SUM
            this.RecountNUM();
            this.RecountSUM();
        };
        
        let result = Buffer.alloc(this.NUM + 4, 0);
        result[0] = this.PRE;
        result[1] = this.FRM;
        result.writeUInt16BE(this.NUM, 2);
        result[4] = this.ADR;
        result[5] = this.SIG;
        result[6] = this.INST;
        let offset = 7;
        if (Buffer.isBuffer(this.DATA)) offset += this.DATA.copy(result, 7, 0);     // buf.copy(target[, targetStart[, sourceStart[, sourceEnd]]])
        result[offset] = this.SUM;
        offset++;
        result[offset] = this.CR;
        return result;
    }

    GetHex = (separator) => {
        let arr = Array.from(this.GetBuffer(false));
        return arr.map(num => "0x"+NumToHex(num)).join(separator || "");
    }
}

class SpinelRx {
    
    buffLength = 0x3FFFF;   // 256 kB
    allowInvalid = false;   // Packets with error also include to the output
    onOk = null;
    onTimeout = null;

    buffWriteIndex = 0;
    buffReadIndex = 0;
    buffer = Buffer.allocUnsafe(this.buffLength);                                      // receiving buffer
    
    /**
     * Processing of received data.
     * @param {Boolean} params.allowInvalid Packets with error also include to the output.
     * @param {Buffer | Array} [params.data] New data to buffer.
     */
    constructor(params) {
        if (Array.isArray(params) || Buffer.isBuffer(params)) {
            this.newRxData(params); 
        } else {
            if (typeof params.allowInvalid === "boolean") this.allowInvalid = params.allowInvalid;
            if (typeof params.onOk === "function") this.onOk = params.onOk;
            if (typeof params.onTimeout === "function") this.onTimeout = params.onTimeout;
            if (Array.isArray(params.data) || Buffer.isBuffer(params.data)) {
                this.newRxData(params.data); 
            }
        }
    }

    resetTimeout(timeoutMs) {
        this.stop = performance.now() + timeoutMs;
        this.timeout = timeoutMs;
    }

    getSpinelPackets(data) {

        let dataLength = data.length;
        let i = 0;
        let out = {packets: [], moveReadIndex: 0};
    
        while ((i + 8) <= dataLength) {
            
            if ((data[i] == 0x2A) && (data[i+1] == 0x61)) {         // header searching (PRE+FRM)
    
                let one = {};
                one.idPRE = i;
                let num = data.readUInt16BE(i+2) - 5;     // calculating data length
    
                if ((i + num + 8 <= dataLength) && (data[i + num + 8] == 0x0D))  {  // is the full length of the packet in the buffer (according to NUM) and is the last byte CR?
        
                    let packet = new PacketSpinel97(false);
                    packet.PRE = data[i];
                    packet.FRM = data[i + 1];
                    packet.NUM = data.readUInt16BE(i+2);
                    packet.ADR = data[i + 4];
                    packet.SIG = data[i + 5];
                    packet.INST = data[i + 6];
                    if (num > 0) {
                        packet.DATA = Buffer.alloc(num, 0x00);
                        data.copy(packet.DATA, 0, i+7, i+7+num);     // buf.copy(target[, targetStart[, sourceStart[, sourceEnd]]])
                    }
                    packet.SUM = data[i + 7 + num];
                    one.idCR = i + 8 + num;
                    packet.CR = data[one.idCR];
                    
                    one.errors = packet.Check();
                    one.valid = !Array.isArray(one.errors);
                    one.packet = {
                        buffer: packet.GetBuffer(false),
                        string: packet.GetHex(","),
                        object: {
                            ADR: packet.ADR,
                            SIG: packet.SIG
                        }
                    }
                    if (packet.INST <= 0x0F) one.packet.object.ACK = packet.INST;
                    else one.packet.object.INST = packet.INST;
                    one.packet.object.DATA = packet.DATA || [];
                    
                    out.packets.push(one);
                    out.moveReadIndex = i + 8 + num + 1;
                    i = i + 8 + num;
    
                }
            }
            i++;
    
        };
        return out;
    
    }

    newRxData(data) {

        if (data === undefined) return false;

        // console.log(`A: New bytes: ${bytes.length}, ${new Date().toISOString()}`, bytes);

        // Add new data to main circle buffer
        let lastWriteIndex;
        for (const byte of data) {
            this.buffer[this.buffWriteIndex] = byte;
            lastWriteIndex = this.buffWriteIndex;
            this.buffWriteIndex++;
            if (this.buffWriteIndex >= this.buffLength) this.buffWriteIndex = 0;
        }
        
        // Copy unprocessed data into one new buffer
        let pieceLength;
        let piece;
        if (this.buffReadIndex < lastWriteIndex) {
            pieceLength = lastWriteIndex + 1 - this.buffReadIndex;
            piece = Buffer.allocUnsafe(pieceLength);
            this.buffer.copy(piece, 0, this.buffReadIndex, lastWriteIndex + 1);     // buf.copy(target[, targetStart[, sourceStart[, sourceEnd]]])
        } else {
            pieceLength = (lastWriteIndex + 1) + (this.buffLength - this.buffReadIndex);
            piece = Buffer.allocUnsafe(pieceLength);
            let bytes = this.buffer.copy(piece, 0, this.buffReadIndex);                    // unprocessed before buffer end
            this.buffer.copy(piece, bytes, 0, lastWriteIndex + 1);   // unprocessed from buffer start
        }
        if (piece.length > (this.buffLength * 0.75)) console.log("Buffer is more than 75% full!");
        if (piece.length < 9) return false;
        
        let allPackets = this.getSpinelPackets(piece);
        if (allPackets.moveReadIndex > 0) {
            if ((this.buffReadIndex + allPackets.moveReadIndex) > this.buffLength) 
                this.buffReadIndex = this.buffReadIndex + allPackets.moveReadIndex - this.buffLength;
            else 
                this.buffReadIndex += allPackets.moveReadIndex;
        };

        let found = 0;
        for (const one of allPackets.packets) {
            if (one.valid) {                                      // packet valid
                // console.log(`B: ${one.packet.SIG}`, this.buffReadIndex);
                if (this.onOk) this.onOk(one.packet, one.errors);
                found++;
            } else if (this.allowInvalid && (Array.isArray(one.errors))) {                          // found invalid packet
                if (this.onOk) this.onOk(one.packet, one.errors);
                found++;
            } else if (one.idPRE > 0) {                           // found potential packet start - delete previous data
                console.log('Found PRE:', one.idPRE);
            }
        }
        if (found > 1) console.log("Packets: ", found);
        return found > 0;

    }

}

module.exports = { PacketSpinel97, SpinelRx, GetSig };